import numpy as np
from matplotlib import pyplot as plt
import seaborn as sns


def large_font_size():
    parameters = {
        "axes.labelsize": 20,
        "axes.titlesize": 20,
        "figure.titlesize": 20,
        "xtick.labelsize": 20,
        "ytick.labelsize": 20,
        "legend.fontsize": 15,
        "legend.title_fontsize": 16,
        "axes.titleweight": "bold",
        "axes.labelweight": "bold",
        "figure.titleweight": "bold",
        "font.weight": "bold",
        "font.sans-serif": "Arial",
    }
    plt.rcParams.update(parameters)


def small_font_size():
    parameters = {
        "axes.labelsize": 6,
        "axes.titlesize": 12,
        "figure.titlesize": 12,
        "xtick.labelsize": 6,
        "ytick.labelsize": 6,
        "legend.fontsize": 8,
        "legend.title_fontsize": 8,
    }
    plt.rcParams.update(parameters)


def no_edge(ax):
    sns.set_style("white")
    ax.spines["right"].set_color("none")
    ax.spines["top"].set_color("none")
    ax.xaxis.set_ticks_position("bottom")
    ax.yaxis.set_ticks_position("left")

    return ax



import numpy as np
import networkx as nx
from networkx.algorithms import bipartite


def create_bipartite_graph(adj_mat, num_nodes_set1, num_nodes_set2, corr_thresh=0.2):
    # Filter adjacency matrix based on the correlation threshold
    adj_mat = np.where(np.abs(adj_mat) >= corr_thresh, adj_mat, 0)

    # Create an empty bipartite graph
    bipartite_graph = nx.Graph()

    # Add nodes to the graph
    bipartite_graph.add_nodes_from(range(1, num_nodes_set1 + 1), bipartite=0)
    bipartite_graph.add_nodes_from(
        range(num_nodes_set1 + 1, num_nodes_set1 + num_nodes_set2 + 1), bipartite=1
    )

    # Add weighted edges between the two sets
    for node_set1 in range(1, num_nodes_set1 + 1):
        for node_set2 in range(num_nodes_set1 + 1, num_nodes_set1 + num_nodes_set2 + 1):
            if adj_mat[node_set1 - 1, node_set2 - 1] == 0:
                continue
            weight = round(adj_mat[node_set1 - 1, node_set2 - 1], 2)
            bipartite_graph.add_weighted_edges_from([(node_set1, node_set2, weight)])

    return bipartite_graph


def create_full_graph(adj_mat, corr_thresh=0.2):
    # Filter adjacency matrix based on the correlation threshold
    adj_mat = np.where(np.abs(adj_mat) >= corr_thresh, adj_mat, 0)

    # Create an empty graph
    full_graph = nx.Graph()

    # Add nodes to the graph
    num_nodes = adj_mat.shape[0]
    full_graph.add_nodes_from(range(1, num_nodes + 1))

    # Add weighted edges
    for i in range(num_nodes):
        for j in range(i + 1, num_nodes):
            if adj_mat[i, j] == 0:
                continue
            weight = round(adj_mat[i, j], 2)
            full_graph.add_weighted_edges_from([(i + 1, j + 1, weight)])

    return full_graph


# TODO: Define the metrics that are not directly provided
# TODO: Check if the small_worldness are correct
def small_worldness(G):
    # calculate the small worldness of the graph
    L = nx.average_shortest_path_length(G)
    C = nx.average_clustering(G)
    n = len(G.nodes())
    p = nx.density(G)
    random_graph = nx.watts_strogatz_graph(n, 2, p)
    L_r = nx.average_shortest_path_length(random_graph)
    C_r = nx.average_clustering(random_graph)
    return (C / C_r) / (L / L_r)


def compute_bipa_graph_metrics(bipartite_graph, nodes_set1):
    metrics = {
        "clustering": bipartite.clustering(bipartite_graph),
        "clustering_coefficient": bipartite.cluster.clustering(bipartite_graph),
        "betweenness_centrality": bipartite.betweenness_centrality(
            bipartite_graph, nodes_set1
        ),
        "degree_centrality": bipartite.degree_centrality(bipartite_graph, nodes_set1),
        "closeness_centrality": bipartite.closeness_centrality(
            bipartite_graph, nodes_set1
        ),
    }
    return metrics


def compute_full_graph_metrics(full_graph):
    metrics = {
        "clustering": nx.clustering(full_graph),
        "clustering_coefficient": nx.average_clustering(full_graph),
        "betweenness_centrality": nx.betweenness_centrality(full_graph),
        "degree_centrality": nx.degree_centrality(full_graph),
        "closeness_centrality": nx.closeness_centrality(full_graph),
    }
    return metrics



from neuroimager.utils.atlas import (
    split_schafer_names,
    combine_atlases,
    combine_probabilistic_atlases,
    filter_rois,
    resample_masks,
)
from neuroimager.utils.fc import (
    flatten_lower_triangular,
    unflatten_lower_triangular,
    filter_fcs,
    extract_bipartite_matrix,
    average_nodes,
)

from neuroimager.utils.ml import (
    evaluate_continuous,
    evaluate_binary,
)

from neuroimager.utils.rbload import (
    rbload_csv,
    rbload_imgs,
)

__all__ = [
    "split_schafer_names",
    "combine_atlases",
    "combine_probabilistic_atlases",
    "resample_masks",
    "filter_rois",
    "flatten_lower_triangular",
    "unflatten_lower_triangular",
    "filter_fcs",
    "extract_bipartite_matrix",
    "average_nodes",
    "evaluate_continuous",
    "evaluate_binary",
    "rbload_csv",
    "rbload_imgs",
]



# Date: 2023-06
# Author: Jianzhang Ni (weitqe@GitHub), UoB & CUHK

import os

import numpy as np
import pandas as pd
from scipy.io import loadmat
from sklearn.decomposition import PCA

from matplotlib import pyplot as plt
import seaborn as sns
from nilearn import plotting


class HmmParser(object):
    def __init__(
        self,
        hmm: str or dict,
        volumes: int,
        subj_num: int,
        sessions: int = 1,
        vpath=None,
        gamma=None,
        xi=None,
        subj_labels=None,
        roi_labels=None,
        auto_parse=True,
        generate_report=True,
        output_dir=None,
    ):
        if isinstance(hmm, str):
            # most simple way to use this class
            self._load_hmm(hmm)
        elif isinstance(hmm, dict):
            # If you already loaded the hmm .mat file
            self.hmm = self.convert_hmm(hmm)
            self.vpath = vpath
            self.gamma = gamma
            self.Xi = xi
        else:
            raise TypeError("hmm must be a file path or a dictionary")
        self.volumes = volumes
        self.subj_num = subj_num
        self.sessions = sessions
        self.roi_labels = roi_labels
        if subj_labels:
            self.subj_labels = subj_labels
        else:
            self.subj_labels = [f"subj{i}" for i in range(subj_num)]
        if self.sessions > 1:
            self.subj_labels = self._generate_subj_labels_with_session()
        if output_dir is None:
            self.output_dir = os.getcwd()
        else:
            self.output_dir = output_dir
        self._check_hmm()
        self.K = int(self.hmm["K"])
        self.states_info = {}
        if auto_parse:
            self.chronnectome = self.parse_chronnectome()
            for i, (mean, conn) in enumerate(zip(self.get_means(), self.get_conns())):
                self.states_info[f"state{i}"] = {"mean": mean, "conn": conn}
            self.P = self.transition_matrix()
        if generate_report:
            self.generate_report()

    """
    *****************************************************************************
    Part 1:  prepare the object
    *****************************************************************************
    """

    def _check_hmm(self):
        # check the all of the above parameters, if one of them is not provided, raise error
        param_dict = {
            "hmm": self.hmm,
            "vpath": self.vpath,
            "gamma": self.gamma,
            "volumes": self.volumes,
            "subj_num": self.subj_num,
            "sessions": self.sessions,
        }
        flat = []
        for key, item in param_dict.items():
            if item is None:
                flat.append(key)
        if len(flat) > 0:
            raise ValueError(f"The following parameters are missing: {flat}")
        if self.subj_num * self.sessions * self.volumes != self.vpath.shape[0]:
            raise ValueError(
                f"The number of volumes in vpath ({self.vpath.shape[0]}) does not match the number of "
                f"volumes in the data ({self.subj_num * self.sessions * self.volumes})"
            )

    def _load_hmm(self, hmm_file: str):
        try:
            hmm = loadmat(hmm_file)
            self.hmm = self.convert_hmm(hmm)
            self.vpath = np.squeeze(hmm["vpath"])
            self.gamma = np.array(hmm["Gamma"])
            self.Xi = np.array(hmm["Xi"])
        except Exception as e:
            print(f"Error loading HMM file: {e}")

    def _generate_subj_labels_with_session(self):
        subject_session_ids = []
        for subj in self.subj_labels:
            for ses in range(1, self.sessions + 1):
                subject_session_id = f"{subj}_s{ses + 1}"
                subject_session_ids.append(subject_session_id)
        return subject_session_ids

    # get hmm parameters
    """
    *****************************************************************************
    Part 2:  prepare the object
    *****************************************************************************
    """

    @staticmethod
    def convert_hmm(hmm):
        """
        ***Only tested for FMRI DATA***
        Convert the `hmm` object derived from the following command of HMM-MAR Toolbox
        `[hmm, Gamma, Xi, vpath] = hmmmar(f,T,options);`
        to a Python Dictionary for further processing
        """
        hmm_dict = {}
        hmm_param = [
            "train",
            "K",
            "prior",
            "Dir_alpha",
            "Pi",
            "Dir2d_alpha",
            "P",
            "state",
        ]
        train_param = [
            "order",
            "verbose",
            "cyc",
            "initrep",
            "initcyc",
            "standardise",
            "inittype",
            "zeromean",
            "covtype",
            "useParallel",
            "repetitions",
            "Fs",
            "K",
            "onpower",
            "leida",
            "embeddedlags",
            "pca",
            "pca_spatial",
            "lowrank",
            "varimax",
            "maxFOth",
            "pcamar",
            "pcapred",
            "filter",
            "detrend",
            "downsample",
            "leakagecorr",
            "sequential",
            "standardise_pc",
            "regularisation",
            "Gamma_constraint",
            "acrosstrial_constrained",
            "plotGamma",
            "id_mixture",
            "episodic",
            "cluster",
            "initcriterion",
            "meancycstop",
            "cycstogoafterevent",
            "initTestSmallerK",
            "stopcriterion",
            "tol",
            "behaviour",
            "tudamonitoring",
            "tuda",
            "distribution",
            "grouping",
            "Pstructure",
            "Pistructure",
            "dropstates",
            "priorcov_rate",
            "timelag",
            "exptimelag",
            "orderoffset",
            "symmetricprior",
            "uniqueAR",
            "S",
            "prior",
            "Sind",
            "orders",
            "maxorder",
            "DirichletDiag",
            "PriorWeightingP",
            "PriorWeightingPi",
            "hmm",
            "fehist",
            "updateObs",
            "updateGamma",
            "updateP",
            "decodeGamma",
            "keepS_W",
            "useMEX",
            "ndim",
            "active",
        ]
        prior_param = ["Dir2d_alpha", "Dir_alpha"]
        omega_param = ["Gam_shape", "Gam_rate"]
        sub_param = {"train": train_param, "prior": prior_param, "Omega": omega_param}

        hmm = hmm["hmm"][0][0]
        for i, param in enumerate(hmm_param):
            data = hmm[i]

            if param in sub_param.keys():
                data = data[0][0]
                tem_dict = {}

                for j, sub in enumerate(sub_param[param]):
                    tem_dict[sub] = data[j]
                hmm_dict[param] = tem_dict
                if param == "train":
                    if tem_dict["covtype"] in ["sharedfull", "uniquefull"]:
                        hmm_param.append("Omega")
                        prior_param.append("Omega")
                        omega_param.append("Gam_irate")
                        unique = True
                    elif tem_dict["covtype"] in ["shareddiag", "uniquediag"]:
                        hmm_param.append("Omega")
                        prior_param.append("Omega")
                        unique = True
                    else:
                        unique = False

            elif param == "state":
                data = data[0]
                tem_dict = {}
                for k in range(len(data)):
                    pr_dict, w_dict = {}, {}
                    state_data = data[k]
                    prior_data = state_data[0][0][0]
                    state_prior_par = ["sigma", "alpha", "Mean"]
                    w_data = state_data[1][0][0]
                    state_w_par = ["Mu_W", "S_W", "iS_W"]

                    for m, sub_par in enumerate(state_prior_par):
                        pr_dict[sub_par] = prior_data[m]
                    for m, sub_par in enumerate(state_w_par):
                        w_dict[sub_par] = w_data[m]
                    if not unique:
                        omega_data = state_data[2][0][0]
                        state_omega_par = [
                            "Gam_rate",
                            "Gam_shape",
                        ]
                        omega_dict = dict()
                        for m, sub_par in enumerate(state_omega_par):
                            omega_dict[sub_par] = omega_data[m]
                        tem_dict[f"state{k + 1}"] = {
                            "prior": pr_dict,
                            "W": w_dict,
                            "Omega": omega_dict,
                        }
                        continue
                    elif unique:
                        tem_dict[f"state{k + 1}"] = {
                            "prior": pr_dict,
                            "W": w_dict,
                        }
                hmm_dict[param] = tem_dict
            else:
                hmm_dict[param] = np.squeeze(data)

        return hmm_dict

    """
    *****************************************************************************
    Part 3:  Calculate chronnectome features
    *****************************************************************************
    """

    # Calculate Vpath features
    def vpath_fo(self, vpath):
        state_fo = dict()
        length = len(vpath)
        k_states = len(np.unique(vpath))
        max_fo = 0
        max_state = 0
        for state in range(1, k_states + 1):
            fo = [np.count_nonzero(vpath == state) / length]
            if fo == np.nan:
                fo = 0
            state_fo[f"state{state}_fo"] = fo
            if state == 1:
                max_fo = fo
                max_state = state
            else:
                if fo > max_fo:
                    max_fo = fo
                    max_state = state
        state_fo["vpath_max_fo"] = max_fo
        return pd.DataFrame.from_dict(state_fo)

    @staticmethod
    def gamma_fo(gamma):
        gamma_fo_array = gamma.sum(axis=0) / gamma.shape[0]
        state_names = [f"state{i + 1}_fo" for i in range(gamma.shape[1])]
        gamma_fo_df = pd.DataFrame(gamma_fo_array, index=state_names).T
        gamma_fo_df["gamma_max_fo"] = np.max(gamma_fo_array)

        return gamma_fo_df

    def vpath_visit(self, vpath):
        dic = dict()
        length = len(vpath)
        k_states = self.K
        for state in range(1, k_states + 1):
            vpath_bool = vpath == state
            visit = 0
            continuous = False
            for i in range(length):
                if not continuous:
                    if vpath_bool[i]:
                        visit += 1
                    else:
                        continue
                try:
                    if vpath_bool[i + 1]:
                        continuous = True
                    else:
                        continuous = False
                except IndexError:
                    pass
            dic[f"state{state}_visits"] = [visit]

        return pd.DataFrame.from_dict(dic)

    @staticmethod
    def vpath_switch(vpath):
        dic = dict()
        length = len(vpath)
        change = 0
        for i in range(length - 1):
            if vpath[i] != vpath[i + 1]:
                change += 1
        dic["switch_rate"] = [change / length]

        return pd.DataFrame.from_dict(dic)

    def vpath_lifetime(self, vpath, mean=True):
        dic = dict()
        length = len(vpath)
        k_states = self.K
        for state in range(1, k_states + 1):
            lifes = []
            vpath_bool = vpath == state
            life = 0
            for i in range(length):
                if vpath_bool[i]:
                    life += 1
                elif (vpath_bool[i] == False) & (life > 0):
                    lifes.append(life)
                    life = 0
            if len(lifes) == 0:
                dic[f"state{state}_lifetime"] = [0]
            elif mean:
                dic[f"state{state}_lifetime"] = [np.mean(lifes)]
            else:
                dic[f"state{state}_lifetime"] = [lifes]

        return pd.DataFrame.from_dict(dic)

    def vpath_interval(self, vpath, mean=True):
        dic = dict()
        length = len(vpath)
        k_states = self.K
        for state in range(1, k_states + 1):
            intervals = []
            vpath_bool = vpath == state
            interval = 0
            for i in range(length):
                if not vpath_bool[i]:
                    interval += 1
                elif (vpath_bool[i] == True) & (interval > 0):
                    intervals.append(interval)
                    interval = 0
            if len(intervals) == 0:
                dic[f"state{state}_interval"] = [0]
            elif mean:
                dic[f"state{state}_interval"] = [np.mean(intervals)]
            else:
                dic[f"state{state}_interval"] = [intervals]

        return pd.DataFrame.from_dict(dic)

    def parse_chronnectome(self, mean=True):
        """
        Parse chronnectome data and compute various visitation metrics for each subject.

        Parameters
        ----------
        mean : bool, controls if the mean of the interval and lifetime metrics are returned or the raw values

        Returns
        -------
        results : dict
            A dictionary containing DataFrames for each visitation metric, with keys corresponding to the metric names
            The metrics include:
            - vpath_fo: Fractional Occupancy
            - vpath_visit: Visitation counts
            - vpath_lifetime: State lifetimes
            - vpath_interval: State intervals
            - vpath_switch: Switching rate

        """
        funcs = [
            self.vpath_fo,
            self.gamma_fo,
            self.vpath_visit,
            self.vpath_switch,
            self.vpath_lifetime,
            self.vpath_interval,
        ]
        results = {}
        for func in funcs:
            dfs = []
            for i in range(self.subj_num):
                subj_vpath = self.vpath[self.volumes * i : self.volumes * (i + 1)]
                subj_gamma = self.gamma[self.volumes * i : self.volumes * (i + 1), :]
                if func.__name__ in ["vpath_lifetime", "vpath_interval"]:
                    subj_df = func(vpath=subj_vpath, mean=mean)
                elif func.__name__ in ["gamma_fo"]:
                    subj_df = func(subj_gamma)
                else:
                    subj_df = func(vpath=subj_vpath)
                try:
                    subj_df.index = [self.subj_labels[i]]
                except ValueError:
                    raise ValueError(
                        "The number of subject labels does not match the number of subjects"
                    )
                dfs.append(subj_df)

            results[func.__name__] = pd.concat(dfs)
        chronnectome_df = pd.concat(
            [
                results["vpath_fo"],
                results["gamma_fo"],
                results["vpath_visit"],
                results["vpath_lifetime"],
                results["vpath_interval"],
                results["vpath_switch"],
            ],
            axis=1,
        )
        return chronnectome_df

    """
    *****************************************************************************
    Part 4:  Calculate state features
    *****************************************************************************
    """

    def get_means(self):
        means = []
        for i in range(self.K):
            mean = np.squeeze(self.hmm["state"][f"state{i + 1}"]["W"]["Mu_W"])
            means.append(mean)

        return means

    def _conn(self, k):
        # calculate cov and corr matrix
        """
        This function only implements the third part of the following matlab function:
        References:
        if do_HMM_pca
            if ~original_space
                warning('Connectivity maps will necessarily be in original space')
            end
            ndim = size(hmm.state(k).W.Mu_W,1);
            covmat = hmm.state(k).W.Mu_W * hmm.state(k).W.Mu_W' + ...
                hmm.Omega.Gam_rate / hmm.Omega.Gam_shape * eye(ndim);
            icovmat = - inv(covmat);
            icovmat = (icovmat ./ repmat(sqrt(abs(diag(icovmat))),1,ndim)) ...
                ./ repmat(sqrt(abs(diag(icovmat)))',ndim,1);
            icovmat(eye(ndim)>0) = 0;
        elseif is_diagonal
            covmat = diag( hmm.state(k).Omega.Gam_rate / (hmm.state(k).Omega.Gam_shape-1) );
            if ~isfield(hmm.state(k).Omega,'Gam_irate')
                hmm.state(k).Omega.Gam_irate = 1 ./ hmm.state(k).Omega.Gam_irate;
            end
            icovmat = diag( hmm.state(k).Omega.Gam_irate * (hmm.state(k).Omega.Gam_shape-1) );
        else
            ndim = length(hmm.state(k).Omega.Gam_rate);
            covmat = hmm.state(k).Omega.Gam_rate / (hmm.state(k).Omega.Gam_shape-ndim-1);
            if ~isfield(hmm.state(k).Omega,'Gam_irate')
                hmm.state(k).Omega.Gam_irate = inv(hmm.state(k).Omega.Gam_rate);
            end
            icovmat = hmm.state(k).Omega.Gam_irate * (hmm.state(k).Omega.Gam_shape-ndim-1);
            icovmat = (icovmat ./ repmat(sqrt(abs(diag(icovmat))),1,ndim)) ...
                ./ repmat(sqrt(abs(diag(icovmat)))',ndim,1);
            icovmat(eye(ndim)>0) = 0;
        end
        """
        if self.hmm["train"]["covtype"] in [
            "sharedfull",
            "uniquefull",
            "shareddiag",
            "uniquediag",
        ]:
            state = self.hmm["Omega"]
        else:
            state = self.hmm["state"][f"state{k}"]["Omega"]

        if self.hmm["train"]["covtype"] in ["diag", "shareddiag", "uniquediag"]:
            cov_mat = np.diag(state["Gam_rate"] / (state["Gam_shape"] - 1))
            if "Gam_irate" not in state.keys():
                state["Gam_irate"] = 1 / state["Gam_rate"]
            icov_mat = np.diag(state["Gam_irate"] * (state["Gam_shape"] - 1))
        else:
            ndim = state["Gam_rate"].shape[0]
            cov_mat = state["Gam_rate"] / (state["Gam_shape"] - ndim - 1)
            if "Gam_irate" not in state.keys():
                state["Gam_irate"] = np.linalg.inv(state["Gam_rate"])
            icov_mat = state["Gam_irate"] * (state["Gam_shape"] - ndim - 1)
            icov_mat = icov_mat / np.sqrt(np.abs(np.diag(icov_mat)))
            icov_mat = icov_mat / np.sqrt(np.abs(np.diag(icov_mat))).T
            np.fill_diagonal(icov_mat, 0)

        def correlation_from_covariance(covariance):
            v = np.sqrt(np.diag(covariance))
            outer_v = np.outer(v, v)
            correlation = covariance / outer_v
            correlation[covariance == 0] = 0
            return correlation

        corr_mat = correlation_from_covariance(cov_mat)

        return cov_mat, icov_mat, corr_mat

    def get_conns(self):
        # loop through k
        conns = []
        for k in range(1, self.K + 1):
            _, _, corr_mat = self._conn(k)
            conns.append(corr_mat)

        return conns

    """
    *****************************************************************************
    Part 5:  Calculate graph features
    *****************************************************************************
    """

    def transition_matrix(self):
        """
        Calculate the transition matrix of the HMM model

        Notes
        This function doesn't consider multiple sessions as the PNAS paper does.
        Since it is just a method to reorder the transition matrix.
        Feel free to use any other clustering method.
        Returns
        -------
        transition_matrix : pd.DataFrame
            The reordered non-stationary transition matrix of the HMM model
        """
        transition_matrix = self.hmm["P"]
        for j in range(self.K):
            transition_matrix[j, j] = 0
            transition_matrix[j, :] = transition_matrix[j, :] / np.sum(
                transition_matrix[j, :]
            )
        state_labels = [f"state{i + 1}" for i in range(self.K)]
        transition_matrix = pd.DataFrame(
            transition_matrix, index=state_labels, columns=state_labels
        )

        # gamma_sub_mean = np.mean(self.gamma, axis=1).squeeze()

        pca = PCA(n_components=1)
        pca1 = pca.fit_transform(self.gamma.T)
        order = np.argsort(pca1, axis=0).ravel()
        transition_matrix = transition_matrix.iloc[order, order]

        return transition_matrix

    def get_graph(self, threshold=0.2):
        import networkx as nx

        if self.P is None:
            self.P = self.transition_matrix()
        graph_data = self.P.values
        graph_data[graph_data < threshold] = 0
        graph_data[graph_data >= threshold] = 1
        graph = nx.DiGraph()  # or DiGraph, MultiGraph, MultiDiGraph, etc
        for node in self.P.columns:
            for other_node in self.P.index:
                if node == other_node:
                    continue
                graph.add_weighted_edges_from(
                    [(node, other_node, self.P.loc[node, other_node])]
                )

        return graph

    """
    *****************************************************************************
    Part 6:  Visualization
    *****************************************************************************
    """

    def plot_means(self, roi_labels=None):
        means = self.get_means()
        fig, axes = plt.subplots(self.K, figsize=(len(means) * 2.5, self.K * 1.2))
        fig.suptitle("Mean activation of each state")
        if roi_labels is None:
            roi_labels = self.roi_labels
        for i in range(self.K):
            ax = axes[i]
            mean = means[i]
            bar = np.max(np.abs(means[i])) * 0.9
            sns.heatmap(
                pd.DataFrame(mean, columns=[f"state{i + 1}"], index=roi_labels).T,
                cmap="RdBu_r",
                vmin=-bar,
                vmax=bar,
                ax=ax,
                cbar=True,
            )
        plt.subplots_adjust(hspace=1, wspace=0.5)

        return fig

    def plot_conns(self):
        conns = self.get_conns()
        dim = conns[0].shape[0]
        if self.K % 2 == 1:
            row = int((self.K // 2) + 1)
        elif self.K % 2 == 0:
            row = int(self.K // 2)
        fig, axes = plt.subplots(row, 2, figsize=(12, 6 * int(self.K / 2)))
        axes = axes.ravel()
        fig.suptitle("Connectivity of each state")
        for i in range(self.K):
            ax = axes[i]
            conn = conns[i]
            plotting.plot_matrix(
                conn,
                axes=ax,
                colorbar=True,
                # vmax=1,
                # vmin=-1,
                reorder=False,
            )
            ax.set_title(f"State {i + 1}")
        plt.subplots_adjust(
            hspace=0.5, wspace=0.5, top=0.9, bottom=0.1, left=0.1, right=0.9
        )

        return fig

    def plot_vpath(self, subj_index: int or list = None):
        if isinstance(subj_index, int):
            subj_index = [subj_index]
        elif isinstance(subj_index, list):
            if not all(isinstance(x, int) for x in subj_index):
                raise ValueError("subj_index must be a list of integers")
        elif subj_index is None:
            subj_index = range(len(self.subj_labels))
        figs = []
        for sub in subj_index:
            vpath = self.vpath[
                self.volumes * sub : self.volumes * (sub + 1)
            ]  # trial type
            durations = np.ones_like(vpath)
            onset = range(vpath.shape[0])
            model_event = pd.DataFrame(
                {"onset": onset, "duration": durations, "trial_type": vpath}
            )
            figs.append(
                plotting.plot_event(
                    model_event,
                )
            )
        if len(figs) == 1:
            return figs[0]
        return figs

    def plot_louvain_community(self, threshold=0.2):
        import networkx as nx
        from sknetwork.clustering import Louvain
        from sknetwork.visualization import svg_graph

        graph = self.get_graph(threshold=threshold)
        sparse_graph = nx.to_scipy_sparse_array(graph)
        # not sure why the sparse array is not working with Louvain, have to use numpy array now.
        numpy_array = sparse_graph.toarray()
        louvain = Louvain()
        labels = louvain.fit_transform(numpy_array)
        labels_array = np.array(labels.todense())
        labels_list = list(np.argmax(labels_array, axis=1))
        names = self.P.columns
        pos = nx.spring_layout(graph)
        fig, ax = plt.subplots(figsize=(self.K, self.K))

        # Set colors for nodes using the labels
        cmap = sns.color_palette("husl", len(labels_list))
        node_colors = [cmap[label] for label in labels_list]

        nx.draw(graph, pos, node_color=node_colors, with_labels=True, ax=ax)
        ax.set_title("Louvain Community Clustering")
        return fig

    """
    *****************************************************************************
    Part 7:  Automation
    *****************************************************************************
    """

    def generate_report(self, threshold=0.2, plot_vpath=True):
        # plot the figures and save them, then write a html file which cites them
        # plot the figures
        out = self.output_dir + f"hmm_{self.K}states_derivatives/"
        if not os.path.exists(out):
            os.makedirs(out)
        states_out = f"{out}states/"
        vpath_out = f"{out}vpath/"
        graph_out = f"{out}graph/"
        if not os.path.exists(states_out):
            os.makedirs(states_out)
        if not os.path.exists(vpath_out):
            os.makedirs(vpath_out)
        if not os.path.exists(graph_out):
            os.makedirs(graph_out)
        means = self.plot_means()
        means.savefig(f"{states_out}means.png")
        plt.close(means)
        conns = self.plot_conns()
        conns.savefig(f"{states_out}conns.png")
        plt.close(conns)
        graph = self.plot_louvain_community(threshold=threshold)
        graph.savefig(f"{graph_out}Louvain.png")
        plt.close(graph)
        if plot_vpath:
            for i, subj_label in enumerate(self.subj_labels):
                subj_vpath = self.plot_vpath(i)
                subj_vpath.savefig(f"{vpath_out}vpath_{subj_label}.png")
                plt.close(subj_vpath)

        with open(f"{out}report.html", "w") as f:
            f.write(
                f"""
                <html>
                <body>
                <h1>Mean activation of each state</h1>
                <img src="{states_out}means.png" alt="means">
                <h1>Connectivity of each state</h1>
                <img src="{states_out}conns.png" alt="conns">
                """
            )
            f.write(
                f"""
                <h1>graph features</h1>
                <img src="{graph_out}Louvain.png" alt="Louvain">
                <h1>State visitation path</h1>

                """
            )
            # loop through subjects and add the images
            try:
                for i, subj_label in enumerate(self.subj_labels):
                    f.write(
                        f"""
                        <h2>Subject {subj_label}</h2>
                        <img src="{vpath_out}vpath_{subj_label}.png" alt="vpath_{subj_label}">
                        """
                    )
            except FileNotFoundError:
                pass
            f.write(
                """
                </body>
                </html>
                """
            )


class HmmModelSelector(object):
    def __init__(
        self,
        models_dir: str,
        krange: list or range,
        rep_num: int,
        volumes: int,
        subj_num: int,
        sessions: int = 1,
        subj_labels: list = None,
        prefix: str = None,
        output_dir: str = None,
    ):
        self.models_dir = models_dir
        self.krange = krange
        self.rep_num = rep_num
        self.volumes = volumes
        self.suj_num = subj_num
        self.sessions = sessions
        self.prefix = prefix
        if subj_labels:
            self.subj_labels = subj_labels
        else:
            self.subj_labels = [f"subj{i}" for i in range(subj_num)]
        if output_dir is None:
            self.output_dir = os.getcwd() + "/"
        else:
            self.output_dir = output_dir

    """
    *****************************************************************************
    Part 1:  Basic Operations
    *****************************************************************************
    """

    def __parse_selected_model(self, hmm_file):
        hmm = HmmParser(
            hmm_file,
            self.volumes,
            self.suj_num,
            self.sessions,
            auto_parse=False,
            generate_report=False,
        )
        return hmm

    @staticmethod
    def __check_file(hmm_file):
        # check if a file exists
        if not os.path.exists(hmm_file):
            raise FileNotFoundError(f"{hmm_file} not found")
        return True

    """
    *****************************************************************************
    Part 2:  Similarity Matrix
    *****************************************************************************
    """

    @staticmethod
    def get_gamma_similarity(gamma1: np.array, gamma2: np.array):
        """
        Mote: This function is recreated from getGammaSimilarity.m of HMM-MAR toolbox.
        Does not support calculation of averaged gamma2, so it has to be a single trail.

        Computes a measure of similarity between two sets of state time courses.
        These can have different number of states, but they must have to be the same length.
        similarity: the sum of joint probabilities under the optimal state alignment
        gamma2_order: optimal state alignment for gamma2 (uses munkres algorithm)
        gamma2_reordered: the second set of state time courses reordered to match gamma1

        """
        # check input
        if gamma1.shape[0] != gamma2.shape[0]:
            raise ValueError(
                "gamma1 and gamma2 must have the same number of time points"
            )
        if len(gamma1.shape) != 2 or len(gamma2.shape) != 2:
            raise ValueError("gamma1 and gamma2 must be 2D arrays")

        T, K = gamma1.shape

        gamma1_0 = gamma1.copy()
        M = np.zeros((K, K))  # cost

        g = gamma2

        K2 = g.shape[1]

        if K < K2:
            gamma1 = np.hstack((gamma1_0, np.zeros((T, K2 - K))))
            K = K2
        elif K > K2:
            g = np.hstack((g, np.zeros((T, K - K2))))

        for k1 in range(K):
            for k2 in range(K):
                M[k1, k2] = (
                    M[k1, k2] + (T - np.sum(np.minimum(gamma1[:, k1], g[:, k2]))) / T
                )
        from munkres import Munkres

        munkres_solver = Munkres()
        gamma2_order = munkres_solver.compute(M.copy())
        cost = sum(M[i][j] for i, j in gamma2_order)

        similarity = K - cost

        gamma2_reordered = gamma2[:, gamma2_order]

        return similarity, gamma2_order, gamma2_reordered

    def calc_simi_matrix(self, state_k):
        import warnings

        gamma = {}
        for rep in range(1, self.rep_num + 1):
            try:
                if self.prefix is None:
                    hmm_file = f"{self.models_dir}k{state_k}_rep{rep}.mat"
                else:
                    hmm_file = f"{self.models_dir}{self.prefix}_k{state_k}_rep{rep}.mat"
                self.__check_file(hmm_file)
                hmm = self.__parse_selected_model(hmm_file)
                gamma[rep] = hmm.gamma
            except FileNotFoundError:
                warnings.warn(
                    f"Rep{rep} for k{state_k} not found, skipping", UserWarning
                )
                continue
        size = len(gamma.keys())
        simi_matrix = np.zeros((size, size))
        for i in range(size):
            for j in range(i, size):
                if i == j:
                    simi_matrix[i, j] = 1
                simi_matrix[i, j], _, _ = self.get_gamma_similarity(
                    gamma[i + 1], gamma[j + 1]
                )
                simi_matrix[j, i] = simi_matrix[i, j]

        return simi_matrix

    @staticmethod
    def __mean_simi(simi_matrix, reps=5):
        return (simi_matrix.sum() - reps) / (reps * (reps - 1))

    def get_mean_simi_df(self):
        mean_simi_df = pd.DataFrame(
            [], index=self.krange, columns=["State", "Mean_simi"]
        )
        for K in self.krange:
            if K == 1:
                mean_simi_df.loc[K, "Mean_simi"] = 1
                mean_simi_df.loc[K, "State"] = f"1 States"
            else:
                simi = self.calc_simi_matrix(K)
                mean_simi_df.loc[K, "Mean_simi"] = self.__mean_simi(simi, simi.shape[0])
                mean_simi_df.loc[K, "State"] = f"{K} States"
        return mean_simi_df

    """
    *****************************************************************************
    Part 3:  MaxFO and Switch rate
    *****************************************************************************
    """

    def __get_subj_chronnectome(self, state_k):
        """
        Get the chronnectome of the model
        """
        vpath_max_fo, gamma_max_fo, switch = {}, {}, {}
        for rep in range(1, self.rep_num + 1):
            try:
                if self.prefix is None:
                    hmm_file = f"{self.models_dir}k{state_k}_rep{rep}.mat"
                else:
                    hmm_file = f"{self.models_dir}{self.prefix}_k{state_k}_rep{rep}.mat"
                self.__check_file(hmm_file)
                hmm = self.__parse_selected_model(hmm_file)
                chronnectome = hmm.parse_chronnectome()
                vpath_max_fo[rep] = chronnectome["vpath_max_fo"].values
                gamma_max_fo[rep] = chronnectome["gamma_max_fo"].values
                switch[rep] = chronnectome["switch_rate"].values
            except FileNotFoundError:
                print(f"Rep{rep} for {state_k} not found, skipping")
                continue
        # convert a dict to dataframe
        vpath_max_fo_df = pd.DataFrame.from_dict(vpath_max_fo)
        gamma_max_fo_df = pd.DataFrame.from_dict(gamma_max_fo)
        switch_df = pd.DataFrame.from_dict(switch)

        return vpath_max_fo_df, gamma_max_fo_df, switch_df

    """
    *****************************************************************************
    Part 4:  Visualization
    *****************************************************************************
    """

    @staticmethod
    def plot_simi_matrix(simi_matrix, **kwargs):
        fig = plt.figure(figsize=(5, 5))
        sns.heatmap(simi_matrix, cmap="jet", vmax=1, vmin=-1, **kwargs)
        plt.title(f"Similarity Matrix")
        return fig

    def __plot_chronnectome(self, data, feature_name, ax=None):
        """
        Plot the chronnectome of the model
        Parameters
        ----------
        data: pd.DataFrame with shape (n_subjects, N_reps)
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=(self.rep_num * 2, 5))
        violin_cmap = sns.color_palette("cool", n_colors=self.rep_num, desat=0.8)
        import warnings

        warnings.filterwarnings("ignore", category=UserWarning)
        sns.violinplot(data=data, inner=None, palette=violin_cmap, ax=ax)
        sns.stripplot(data=data, jitter=True, size=3, ax=ax)
        ax.set_ylim(0, 1)
        ax.set_xlabel("Repetition")
        ax.set_ylabel(f"{feature_name}")

        return ax

    def plot_chronnectome(self, state_k):
        from neuroimager.plotting.styler import no_edge

        vpath_max_fo, gamma_max_fo, switch_df = self.__get_subj_chronnectome(state_k)
        fig, axes = plt.subplots(1, 3, figsize=(20, 10))
        self.__plot_chronnectome(vpath_max_fo, "vpath_max_fo", no_edge(axes[0]))
        self.__plot_chronnectome(gamma_max_fo, "gamma_max_fo", no_edge(axes[1]))
        self.__plot_chronnectome(switch_df, "switch", no_edge(axes[2]))
        fig.suptitle(f"Chronnectome for {state_k} states")

        return fig

    def plot_mean_simi(self):
        mean_simi_df = self.get_mean_simi_df()
        fig, ax = plt.subplots(1, 1, figsize=(10, 5))
        from neuroimager.plotting import no_edge

        ax = no_edge(ax)
        fig.suptitle("Mean state similarity")
        sns.barplot(x="State", y="Mean_simi", data=mean_simi_df, color="gray", ax=ax)
        ax.set_xlabel("State")
        ax.set_ylabel("Mean similarity")
        ax.set_ylim(0.4, 1)
        return fig

    """
    *****************************************************************************
    Part 5:  Automation
    *****************************************************************************
    """

    def auto_parse(self):
        out = self.output_dir + f"model_selection/"
        if not os.path.exists(out):
            os.makedirs(out)
        simi_out = f"{out}simi/"
        if not os.path.exists(simi_out):
            os.makedirs(simi_out)
        chrono_out = f"{out}chrono/"
        if not os.path.exists(chrono_out):
            os.makedirs(chrono_out)
        mean_simi_df_fig = self.plot_mean_simi()
        mean_simi_df_fig.savefig(f"{simi_out}mean_simi.png")
        plt.close(mean_simi_df_fig)
        for state_k in self.krange:
            simi_matrix = self.calc_simi_matrix(state_k)
            fig = self.plot_simi_matrix(simi_matrix)
            fig.savefig(f"{simi_out}k{state_k}.png")
            plt.close(fig)
            vpath_max_fo, gamma_max_fo, switch = self.__get_subj_chronnectome(state_k)
            fig = self.plot_chronnectome(state_k)
            fig.savefig(f"{chrono_out}k{state_k}.png")
            plt.close(fig)

        return

    def generate_report(self):
        self.auto_parse()
        out = self.output_dir + f"model_selection/"
        if not os.path.exists(out):
            os.makedirs(out)
        html_file = f"{out}report.html"
        with open(html_file, "w") as f:
            f.write("<html><body>\n")
            f.write("<h1>Model Selection</h1>\n")
            f.write("<h2>Mean Similarity</h2>\n")
            f.write("<img src='simi/mean_simi.png'>\n")
            f.write("<h2>Similarity Matrix</h2>\n")
            for state_k in self.krange:
                f.write(f"<h3>{state_k} States</h3>\n")
                f.write(f"<img src='simi/k{state_k}.png'>\n")
            f.write("<h2>Chronnectome</h2>\n")
            for state_k in self.krange:
                f.write(f"<h3>{state_k} States</h3>\n")
                f.write(f"<img src='chrono/k{state_k}.png'>\n")
            f.write("</body></html>\n")

        return



import re
import pandas as pd
import numpy as np
import nibabel as nib
from nilearn.image import index_img, load_img, resample_to_img
from neuroimager.utils.rbload import rbload_imgs


def symmetrize_image_data(image_data, round: int or False = False):
    mid_point = image_data.shape[0] // 2
    left_hemisphere = image_data[:mid_point, :, :]
    right_hemisphere = image_data[mid_point:, :, :]
    right_hemisphere_flipped = np.flip(right_hemisphere, axis=0)
    symmetrized_data = (left_hemisphere + right_hemisphere_flipped) / 2
    full_symmetrized_data = np.concatenate(
        (symmetrized_data, np.flip(symmetrized_data, axis=0)), axis=0
    )

    return full_symmetrized_data


def symmetrize_image_nifti(input_file, output_paths: str or list = None):
    """
    Symmetrize a NIfTI image across the mid-sagittal plane.
    Returns a NIfTI image object or a list of NIfTI image objects.
    The images are saved if output_paths is specified.

    """
    nifti_images = rbload_imgs(input_file)
    save = False
    if isinstance(output_paths, str):
        output_paths = [output_paths]
        save = True
    elif isinstance(output_paths, list):
        save = True
    if len(output_paths) != len(nifti_images):
        raise ValueError(
            "The number of output paths must match the number of input images."
        )
    symmetrized_nifti_images = []
    for i, nifti_image in enumerate(nifti_images):
        image_data = nifti_image.get_fdata()
        symmetrized_data = symmetrize_image_data(image_data)
        symmetrized_nifti_image = nib.Nifti1Image(
            symmetrized_data, nifti_image.affine, nifti_image.header
        )
        if save:
            nib.save(symmetrized_nifti_image, output_paths[i])
        if len(nifti_images) == 1:
            return symmetrized_nifti_image
        symmetrized_nifti_images.append(symmetrized_nifti_image)

    return symmetrized_nifti_images


def resample_masks(source_img_name, target_img_name, interpolation="continuous"):
    target_img = load_img(target_img_name)
    source_img = load_img(source_img_name)
    res_img = resample_to_img(
        source_img,
        target_img,
        interpolation=interpolation,
        copy=True,
        order="F",
        clip=False,
        fill_value=0,
        force_resample=False,
    )

    if interpolation != "nearest":
        data = res_img.get_fdata()
        data_rounded = data.round().astype(int)
        res_img_rounded = nib.Nifti1Image(data_rounded, res_img.affine, res_img.header)

        return res_img_rounded
    else:
        return res_img


# TODO: Add support for nilean atlas objects
def combine_atlases(atlases: list, output_path: str, return_combined_atlas=True):
    """
    This function combines multiple atlas files (in NIfTI format) into a single atlas file.
    Prioritize atlases in the atlas_paths list by placing them at the front.

    Parameters:
    atlas_paths (list): A list of file paths or nib.nifti1.Nifti1Image to the atlas files to be combined.
            THE ORDER OF THIS LIST MATTERS. As the function will assign the first
            non-zero value to a voxel if there is an overlap.
    output_path (str): The file path where the combined atlas file will be saved.
    return_combined_atlas (bool): If True, the combined atlas will be returned.

    Returns:
    combined_atlas. The combined atlas is returned if return_combined_atlas is True.

    Notes:
    - The input atlas files should have the same affine matrix and dimensions.
    - The function assumes that the input atlases have non-overlapping labels.
    - The combined atlas file will be saved in NIfTI format (.nii.gz).

    Example:
    atlas_path = './assets/masks/'
    files = os.listdir(atlas_path)
    atlas_paths = [os.path.join(atlas_path, file) for file in files]
    output_path = "./assets/output/combined_atlas.nii.gz"
    combine_atlases(atlas_paths, output_path)
    """
    print(
        """Prioritize atlases in the atlas_paths list by placing them at the front.
        The function will assign the first non-zero value to a voxel if there is an overlap."""
    )
    combined_data = None
    max_label = 0
    affine = None
    header = None

    for atlas_file in atlases:
        if isinstance(atlas_file, str):
            atlas = nib.load(atlas_file)
        elif isinstance(atlas_file, nib.nifti1.Nifti1Image):
            atlas = atlas_file
        else:
            raise TypeError(
                "atlas_paths must be a list of file paths or nib.nifti1.Nifti1Image"
            )
        if affine is None:
            affine = atlas.affine
            header = atlas.header
        else:
            if not (atlas.affine == affine).all():
                print(f"Error: Affine matrix mismatch in {atlases}")

        atlas_data = atlas.get_fdata().astype(np.int16)
        if combined_data is None:
            combined_data = atlas_data
        else:
            if atlas_data.shape != combined_data.shape:
                print(f"Error: Dimension mismatch in {atlases}")

            atlas_data[atlas_data > 0] += max_label
            combined_data = np.where(combined_data == 0, atlas_data, combined_data)
        max_label = int(combined_data.max())

    combined_atlas = nib.Nifti1Image(combined_data, affine, header)
    nib.save(combined_atlas, output_path)
    return combined_atlas


def combine_probabilistic_atlases(atlases: list, output_path: str, thresh=0.25):
    """
    This function combines multiple probabilistic atlases into a single atlas by selecting the region with the highest probability for each voxel. The resulting atlas is saved to the specified output path.
    The atlas files must be 4D or 3D probabilistic atlas. The order of the atlases determines the label in the output file.

    Parameters:
    atlas_paths (list): A list of file paths or nib.nifti1.Nifti1Image to the probabilistic atlases to be combined.
            Works for a single file.
    output_path (str): The file path where the combined atlas will be saved.
    thresh (float, optional): A threshold value between 0 and 1 used to create a mask for voxels with low probabilities.Default is 0.25.
        Please Note this threshold is relative to the combined probability not individual atlas probability.

    Returns:
    combined_atlas (nibabel.nifti1.Nifti1Image): The combined atlas as a Nifti1Image object.

    Raises:
    Prints an error message if there is an affine matrix mismatch or dimension mismatch in the input atlases.

    Example:
    combined_atlas = combine_probabilistic_atlases(["atlas1.nii.gz", "atlas2.nii.gz"], "combined_atlas.nii.gz", thresh=0.5)
    """
    combined_data = None
    affine = None
    header = None

    # Check if atlas_paths is a list
    if not isinstance(atlases, list):
        raise TypeError("atlas_paths must be a list even if you have only one file")

    # Check if output_path is a string
    if not isinstance(output_path, str):
        raise TypeError("output_path must be a string")

    for atlas_file in atlases:
        if isinstance(atlas_file, str):
            atlas = nib.load(atlas_file)
        elif isinstance(atlas_file, nib.nifti1.Nifti1Image):
            atlas = atlas_file
        else:
            raise TypeError(
                "atlas_paths must be a list of file paths or nib.nifti1.Nifti1Image"
            )

        if affine is None:
            affine = atlas.affine
            header = atlas.header
        else:
            if not (atlas.affine == affine).all():
                print(f"Error: Affine matrix mismatch in {atlas_file}")

        atlas_data = atlas.get_fdata()

        if atlas_data.ndim == 3:
            # make a 3d prob atlas to 4d prob atlas
            atlas_data = np.expand_dims(atlas_data, axis=-1)
        elif atlas_data.ndim == 4:
            pass
        else:
            print(f"Error: Dimension mismatch in {atlas_file}")

        # make sure all the atlas is on the same scale
        # e.g. some range from 1-100, some range from 0-1
        atlas_data = atlas_data / atlas_data.max()

        if combined_data is None:
            combined_data = atlas_data
        else:
            # concat all the files to one 4d array
            combined_data = np.concatenate((combined_data, atlas_data), axis=3)

    # dim reduce the mask to 3d by summing up the 4th dimension
    dim_red_data = np.sum(combined_data, axis=3)
    mask = dim_red_data < thresh
    # Select the region with the highest probability for each voxel
    max_prob_region_atlas = np.argmax(combined_data, axis=3)
    max_prob_region_atlas += 1  # Add 1 to avoid 0 indexing
    # Apply the mask to the atlas
    max_prob_region_atlas[mask] = 0

    combined_atlas = nib.Nifti1Image(max_prob_region_atlas, affine, header)
    nib.save(combined_atlas, output_path)

    return combined_atlas


def filter_rois(
    atlas: str or nib.nifti1.Nifti1Image,
    rois: list,
    output_path: str,
    remove: bool = True,
):
    """
    This function extracts or exclude specified regions of interest (ROIs) from a given atlas image and saves it.
    This function uses 0 indexing (starts from first ROI) for both 3D and 4D atlases.

    Parameters:
    atlas (str or nib.nifti1.Nifti1Image): The input atlas image, either as a file path string or a nibabel image object.
    output_path (str): The file path where the resulting image will be saved.
    rois (list of int): A list of ROI indices to include or exclude.
    remove (bool, optional): If True, the specified ROIS will be excluded otherwise will be included. Default is True.

    Returns:
    new_atlas_img (nib.nifti1.Nifti1Image): The resulting image with the specified ROIs included or removed.

    Raises:
    ValueError: If the input 'atlas' is not a string or a nibabel image object.
    """
    if isinstance(atlas, str):
        atlas_img = nib.load(atlas)
    elif isinstance(atlas, nib.nifti1.Nifti1Image):
        atlas_img = atlas
    else:
        raise ValueError("atlas must be a string or a nibabel image")

    # Turn the specified ROIS into include list
    if remove:
        # Create a list of all ROI indices
        num_rois = atlas_img.shape[-1]
        all_roi_indices = list(range(num_rois))
        # Remove the specified ROIs from the list
        for roi in rois:
            all_roi_indices.remove(roi)
    elif not remove:
        all_roi_indices = rois
    else:
        raise ValueError("remove must be a boolean")
    all_roi_indices.sort()

    dims = len(atlas_img.shape)
    if dims == 4:
        # Create a new 4D image with only the specified ROIs included
        new_atlas_img = index_img(atlas_img, all_roi_indices)
        nib.save(new_atlas_img, output_path)
    elif dims == 3:
        # Create a boolean mask with the same shape as atlas_data
        atlas_data = atlas_img.get_fdata()
        mask = np.isin(atlas_data, all_roi_indices)
        new_atlas_data = np.where(mask, atlas_data, 0)
        new_atlas_img = nib.Nifti1Image(new_atlas_data, atlas_img.affine)
        nib.save(new_atlas_img, output_path)

    return new_atlas_img


def split_schafer_names(schafer_labels: str or pd.DataFrame):
    """
    This function reads a CSV file containing Schafer parcellation data and extracts the hemisphere, network, and subregion
    information from the ROI names. The extracted information is added as new columns to the DataFrame and the updated
    DataFrame is returned.

    Args:
        schafer_labels (str or pd.DataFrame): The path to the CSV file containing Schafer parcellation data.

    Returns:
        pd.DataFrame: A DataFrame containing the original Schafer parcellation data with additional columns for
                      'Hemisphere', 'Network', and 'Subregion' extracted from the ROI names.
    """
    if isinstance(schafer_labels, str):
        schafer = pd.read_csv(schafer_labels)
    elif isinstance(schafer_labels, pd.DataFrame):
        schafer = schafer_labels
    regex_pattern = r"(RH|LH)_([A-Za-z]+)(?:_([A-Za-z]+))?(?:_[A-Za-z]+)*"

    def parse_roi_name(roi_name):
        match = re.search(regex_pattern, roi_name)
        if match:
            return match.group(1), match.group(2), match.group(3)
        return None, None, None

    schafer[["Hemisphere", "Network", "Subregion"]] = (
        schafer["ROI Name"].apply(parse_roi_name).apply(pd.Series)
    )

    return schafer



import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt


# define a permutation_test function
# p value here is from two-tailed analysis
def perm_ttest_ind(g1, g2, n_reps=5000, plot=True):
    """
    params:
            g1: array of group1
            g2: array of group2
            n_reps: the number of permutation
            plot: whether to plot the permutation distribution
    returns:
            perm_p: two-tailed p value of permutation test
            obs_diff: observed mean difference
            perm_diff: np.array. Contains permuted difference.
    """
    obs_diff = np.mean(g1) - np.mean(g2)
    data = np.append(g1, g2)
    g1_len = len(g1)
    perm_diff = np.zeros(n_reps)
    for i in range(n_reps):
        perm_data = np.random.permutation(data)
        perm_g1 = perm_data[:g1_len]
        perm_g2 = perm_data[g1_len:]
        perm_diff[i] = perm_g1.mean() - perm_g2.mean()
    # two-tailed p
    perm_p = np.count_nonzero(abs(perm_diff) > abs(obs_diff)) / n_reps
    if plot:
        perm_plot(perm_p, obs_diff, perm_diff, arrow_shift=(0.05, 5))
    return perm_p, obs_diff, perm_diff


# plot the result of permutation test
def perm_plot(perm_p, observed, perm_array, arrow_shift=(0.05, 5)):
    """
    params:
        output from function permutation_test()
        arrow_shift: the start position of the arrow relative to the observed value
    """
    sns.distplot(perm_array)
    plt.plot(observed, 0, "ro")
    plt.annotate(
        "p={}".format(perm_p),
        ha="center",
        va="bottom",
        xytext=(observed + arrow_shift[0], arrow_shift[1]),
        xy=(observed, 0),
        arrowprops={"facecolor": "red", "shrink": 0.05},
    )
    plt.show()



from sklearn.linear_model import LinearRegression


def rm_covars_linear(features, covariates):
    """
    Remove multiple covariates from the feature set with linear regression.

    Parameters
    ----------
    features : 2D numpy array, shape (n_samples, n_features_X)
        Set of features.
    covariates : 2D numpy array, shape (n_samples, n_covariates)
        Covariates to regress out.

    Returns
    -------
    new_features : 2D numpy array, shape (n_samples, n_features_X)
        Residuals of the features after regressing out the covariates.
    """
    model = LinearRegression()
    model.fit(covariates, features)
    new_features = features - model.predict(covariates)

    return new_features



from setuptools import setup, find_packages
import os

current_dir = os.path.abspath(os.path.dirname(__file__))
# Read the contents of your README file
with open(os.path.join(current_dir, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="neuroimager",
    version="0.0.9",
    description="A collection of utilities used for MRI data analysis",
    author="Wetiqe",
    author_email="jzni132134@gmail.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Wetiqe/neuroimager",
    packages=find_packages(),
    install_requires=[
        "matplotlib",
        "numpy",
        "pandas==1.5.3",
        "scipy==1.10.1",
        "pingouin",
        "seaborn",
        "sklearn",
        "nilearn",
        "networkx",
        "munkres",
        "scikit-network",
        "pybids",
        "tqdm",
    ],
    extras_require={"torch": ["torch>=1.0.0"]},
    classifiers=[
        "Development Status :: 1 - Planning",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



import os
from tqdm import tqdm
from typing import Callable, List, Dict
from neuroimager.utils.rbload import rbload_csv, rbload_imgs

import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from numpy import ndarray

from sklearn.base import BaseEstimator, TransformerMixin
import nibabel as nib
import bids.layout.models

from nilearn.glm.first_level import FirstLevelModel, check_design_matrix
from nilearn.glm.second_level import SecondLevelModel, non_parametric_inference
from nilearn.glm import threshold_stats_img
from nilearn.image import clean_img
from nilearn.plotting import plot_design_matrix, plot_contrast_matrix
from nilearn.plotting import plot_stat_map, view_img_on_surf, plot_glass_brain
from nilearn.reporting import get_clusters_table


# TODO: Check the input arguments for better debugging
class TaskFmri(BaseEstimator, TransformerMixin, object):
    def __init__(
        self,
        tr: int,
        contrasts: list or dict,
        out_dir: str,
        generate_report: bool = False,
    ):
        self.tr = tr
        self.contrasts = contrasts
        self.out_dir = out_dir
        self.generate_report = generate_report

    def plot_design_matrix(self, design_matrix, prefix="", out_dir=None):
        if out_dir is None:
            out_dir = self.out_dir + "GLM/"
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        plot_design_matrix(design_matrix)
        name = "design_matrix.png"
        if prefix:
            name = prefix + "_" + name
        plt.savefig(out_dir + name)
        plt.close()

    def plot_contrast_matrix(self, prefix="", out_dir=None):
        if out_dir is None:
            out_dir = self.out_dir + "GLM/"
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        if isinstance(self.contrasts, dict):
            for contrast_name, contrast_matrix in self.contrasts.items():
                plot_contrast_matrix(contrast_matrix)
                name = f"{contrast_name}_contrast_matrix.png"
                if prefix:
                    name = prefix + "_" + name
                plt.savefig(out_dir + name)
                plt.close()
        elif isinstance(self.contrasts, list):
            if isinstance(self.contrasts[0], str):
                raise TypeError("Can only plot matrix not contrast name")
            else:
                raise NotImplementedError(
                    "Only supports the plotting of contrast matrix"
                )

    def loop_all_contrasts_parametric(self, fmri_glm, output_prefix):
        return_dict = dict()
        if isinstance(self.contrasts, dict):
            conditions = self.contrasts.keys()
            con_matrix_all = self.contrasts.values()
        elif isinstance(self.contrasts, list):
            conditions = self.contrasts
            con_matrix_all = [None for i in range(len(conditions))]
        else:
            raise TypeError(
                "Contrasts must be a list of conditions or a dictionary of conditions and contrast matrices"
            )
        for condition, con_matrix in zip(conditions, con_matrix_all):
            effect_fname = f"{output_prefix}_{condition}_effect.nii.gz"
            z_fname = f"{output_prefix}_{condition}_z.nii.gz"
            p_value_fname = f"{output_prefix}_{condition}_p.nii.gz"
            if con_matrix is None:
                contrast = condition
            else:
                contrast = con_matrix
            fmri_glm.compute_contrast(contrast, output_type="effect_size").to_filename(
                os.path.join(self.out_dir, effect_fname)
            )

            fmri_glm.compute_contrast(contrast, output_type="z_score").to_filename(
                os.path.join(self.out_dir, z_fname)
            )

            fmri_glm.compute_contrast(contrast, output_type="p_value").to_filename(
                os.path.join(self.out_dir, p_value_fname)
            )
            return_dict[condition] = {
                "effect": self.out_dir + effect_fname,
                "z": self.out_dir + z_fname,
                "p": self.out_dir + p_value_fname,
            }

        return return_dict

    def generate_nilearn_report(
        self,
        fitted_glm: FirstLevelModel or SecondLevelModel,
        contrasts: Dict[str, ndarray] or str or List[str] or ndarray or List[ndarray],
        title=None,
        bg_img="MNI152TEMPLATE",
        threshold=1.96,
        alpha=0.05,
        cluster_threshold=10,
        height_control="fpr",
        min_distance=8.0,
        plot_type="glass",
        display_mode="lyrz",
        report_dims=(1600, 800),
    ):
        html = fitted_glm.generate_report(
            contrasts,
            title=title,
            bg_img=bg_img,
            threshold=threshold,
            alpha=alpha,
            cluster_threshold=cluster_threshold,
            height_control=height_control,
            min_distance=min_distance,
            plot_type=plot_type,
            display_mode=display_mode,
            report_dims=report_dims,
        )
        with open(self.out_dir + f"{title}report.html", "w") as f:
            f.write(html)

    def plot_single_stat(
        self, stat_map, out_name=None, plot="3D", plot_kwargs=None, save=True
    ):
        if plot_kwargs is None and plot == "3D":
            plot_kwargs = {
                "threshold": 3.0,
                "display_mode": "ortho",
                "black_bg": False,
                "title": out_name,
            }
        elif plot_kwargs is None and plot == "glass":
            plot_kwargs = {
                "threshold": 3.0,
                "colorbar": "False",
                "black_bg": False,
                "plot_abs": False,
                "display_mode": "z",
            }
        elif plot_kwargs is None and plot == "surface":
            plot_kwargs = {
                "threshold": 3.0,
                "colorbar": "False",
                "black_bg": False,
                "plot_abs": False,
                "display_mode": "z",
            }

        if plot == "3D":
            plot_stat_map(
                stat_map,
                **plot_kwargs,
            )
        elif plot == "glass":
            plot_glass_brain(
                stat_map,
                **plot_kwargs,
            )
        elif plot == "surface":
            view_img_on_surf(
                stat_map,
                **plot_kwargs,
            )

        if save:
            plt.savefig(os.path.join(self.out_dir, f"{out_name}_zmap.png"))

            # view.save_as_html(
            #     os.path.join(out_dir, f"group_{condition}_{contrast}_zmap.html")
            # )
        plt.close()


class FirstLevelPipe(TaskFmri):
    def __init__(
        self,
        tr: float or int,
        contrasts: List[str] or dict,
        out_dir: str,
        imgs: List[bids.layout.models.BIDSImageFile] = None,
        subj_ids: List[str] = None,
        confounds: List[str or pd.DataFrame] = None,
        confound_items: List[str] = None,
        events: List[str or pd.DataFrame] = None,
        prep_func: str or Callable or None = None,
        generate_report: bool = False,
        first_level_kwargs: dict = None,
    ):
        """

        Note: The imgs, confounds, events can be empty here, in that case they must be specified in the fit method.
        """
        super().__init__(tr, contrasts, out_dir, generate_report)
        self.out_dir = out_dir
        self.tr = tr
        self.imgs = imgs
        self.subj_ids = subj_ids
        self.confounds = confounds
        self.confound_items = confound_items
        self.events = events
        self.prep_func = prep_func
        self.first_level_kwargs = {
            "t_r": self.tr,
            "noise_model": "ar1",
            "hrf_model": "spm",
            "drift_model": "cosine",
            "high_pass": 1.0 / 160,
            "signal_scaling": False,
            "minimize_memory": True,
        }
        self.first_level_kwargs.update(first_level_kwargs)
        self.to_second_level = {}
        self.__prep_input()

    def __prep_input(self):
        os.makedirs(self.out_dir, exist_ok=True)
        if self.imgs is not None:
            if self.subj_ids is not None:
                if len(self.imgs) != len(self.subj_ids):
                    raise ValueError(
                        "The number of images and subject ids must be equal"
                    )
            if self.confounds is None:
                self.confounds = [None] * len(self.imgs)
        if self.confounds is not None:
            if self.confound_items is None:
                self.confound_items = [
                    "white_matter",
                    "framewise_displacement",
                    "trans_x",
                    "trans_y",
                    "trans_z",
                    "rot_x",
                    "rot_y",
                    "rot_z",
                ]
        if isinstance(self.contrasts, dict):
            for key, value in self.contrasts.items():
                self.to_second_level[key] = []
        elif isinstance(self.contrasts, list):
            for contrast in self.contrasts:
                self.to_second_level[contrast] = []
        else:
            raise ValueError("Contrasts must be either dict or list")

    def process_single_subject(
        self,
        img: nib.nifti1.Nifti1Image,
        confound: pd.DataFrame,
        event: pd.DataFrame,
        out_prefix: str,
        plot_design: bool = True,
    ):
        event = event[["trial_type", "onset", "duration"]]
        first_level_model = FirstLevelModel(**self.first_level_kwargs)
        first_level_model = first_level_model.fit(
            img, events=event, confounds=confound
        )  # fit the model
        if self.generate_report:
            self.generate_nilearn_report(first_level_model, self.contrasts, out_prefix)
        design_matrix = first_level_model.design_matrices_[0]
        design_validity = self.__check_design(design_matrix)
        if design_validity == "valid":
            if plot_design:
                self.plot_design_matrix(design_matrix, out_prefix)
        else:
            print(f"Design matrix for {out_prefix} is {design_validity}")
        del img  # release the memory
        subj_results = self.loop_all_contrasts_parametric(first_level_model, out_prefix)

        return subj_results

    def loop_all_subjects(
        self,
        out_prefixes: List[str] = None,
    ):
        """
        Conduct first level analysis for all subjects
        Parameters
        ----------
        out_prefixes: List[str] By default, the subject id is used as the prefix for the output files.
            If the input images are not bids, then the prefix is sub-0, sub-1, etc. Or it will be the subject id if specified.
            If out_prefixes is specified, then it will be used as the prefix for the output files.
            out_prefixes > self.subj_ids > img.get_entities()["subject"] > sub-0, sub-1, etc.

        Returns
        -------

        """
        i = 0
        for img, confound_name, event_name in tqdm(
            zip(self.imgs, self.confounds, self.events)
        ):
            if isinstance(img, bids.layout.models.BIDSImageFile):
                out_prefix = img.get_entities()["subject"]
                img = img.get_image()
            else:
                out_prefix = f"sub-{i}"
                img = rbload_imgs(img)[0]
            if self.subj_ids is not None:
                out_prefix = self.subj_ids[i]
            if out_prefixes is not None:
                out_prefix = out_prefixes[i]
            tqdm.write(f"Processing subject {out_prefix}")
            if isinstance(self.prep_func, Callable):
                img = self.prep_func(img)
            elif self.prep_func == "default":
                img = self.prep_img(img)
            elif isinstance(self.prep_func, str):
                img = self.prep_img(img)
            elif self.prep_func is None:
                pass
            if confound_name is None:
                confound = None
            else:
                confound = rbload_csv(confound_name)[self.confound_items]
            event = rbload_csv(event_name)
            subj_results = self.process_single_subject(img, confound, event, out_prefix)
            for contrast, imgs in subj_results.items():
                self.to_second_level[contrast].append(imgs["z"])
            i += 1

    def prep_img(self, img, **kwargs):
        default_kwargs = {
            "smoothing_fwhm": 6.0,
            "standardize": True,
            "detrend": True,
            "low_pass": 0.1,
            "high_pass": 0.01,
            "confounds": None,
        }
        default_kwargs.update(kwargs)
        preprocessed_img = clean_img(img, t_r=self.tr, **default_kwargs)
        return preprocessed_img

    @staticmethod
    def __check_design(design):
        # TODO: check if design is valid
        frame_times, matrix, names = check_design_matrix(design)
        return "valid"

    def fit(self, X, y=None):
        """
        Parameters
        ----------
        X: X is expected to be a tuple containing (imgs, confounds,confound_items, events)
            confound_items is meaningful only if confounds is not None
        y: For consistency with scikit-learn API

        Returns
        -------
        self
        """
        # You may need to modify this depending on your specific input structure
        imgs, confounds, confound_items, events = X

        self.imgs = imgs
        self.events = events
        self.confound_items = confound_items
        self.confounds = confounds
        self.__prep_input()
        self.loop_all_subjects()

        return self

    def transform(self, X, y=None):
        # You can return the `to_second_level` attribute here,
        # as it contains the results of first-level analysis
        return self.to_second_level


class HigherLevelPipe(TaskFmri):
    def __init__(
        self,
        tr: float or int,
        design_matrix: pd.DataFrame,
        contrasts: List[str] or dict,
        out_dir: str,
        stat_maps: dict = None,
        non_parametric: bool = False,
        generate_report: bool = True,
        stat_map_masks: list or dict = None,
        higher_level_kwargs: dict = None,
    ):
        """
        Higher level analysis pipeline

        Note: stat_maps is a dict containing the first level results, it can be empty when initializing the class
        In that case it must be specified in the fit method

        Parameters
        ----------
        tr
        design_matrix
        contrasts
        out_dir
        stat_maps
        non_parametric
        higher_level_kwargs
        """
        super().__init__(tr, contrasts, out_dir, generate_report)
        self.out_dir = out_dir
        self.stat_maps = stat_maps
        self.design_matrix = design_matrix
        self.higher_results = {}
        self.nonparametric = non_parametric
        self.stat_map_masks = stat_map_masks
        self.model_kwargs = {}
        self.model_kwargs.update(higher_level_kwargs)
        self.__prep_input()

    def __prep_input(self):
        os.makedirs(self.out_dir, exist_ok=True)
        if isinstance(self.contrasts, dict):
            for key, value in self.contrasts.items():
                self.higher_results[key] = []
        elif isinstance(self.contrasts, list):
            for contrast in self.contrasts:
                self.higher_results[contrast] = []
        else:
            raise ValueError("Contrasts must be either dict or list")

    def process_single_1level_contrast(self, stat_maps, out_prefix, plot=True):
        if self.nonparametric:
            higher_level_result = self.loop_all_2level_contrasts_nonparametric(
                stat_maps, out_prefix
            )
        else:
            second_level_model = SecondLevelModel(**self.model_kwargs)
            second_level_model = second_level_model.fit(
                stat_maps, design_matrix=self.design_matrix
            )
            if self.generate_report:
                self.generate_nilearn_report(
                    second_level_model, self.contrasts, out_prefix
                )
            del stat_maps
            higher_level_result = self.loop_all_2level_contrasts_parametric(
                second_level_model, out_prefix
            )
        if plot:
            self.plot_higher_level_result(higher_level_result, out_prefix)

        return higher_level_result

    def plot_higher_level_result(self, higher_level_result, out_prefix):
        # TODO: Update docstrings.
        # TODO: add **kwargs support for all plotting functions
        for contrast_1level, returned_dict in higher_level_result.items():
            if self.nonparametric:
                if isinstance(returned_dict, dict):
                    self.plot_cluster_nonparametric(
                        returned_dict, out_prefix + "_" + contrast_1level
                    )
                else:
                    img = returned_dict
                    fig, ax = plt.subplots(figsize=(10, 5))
                    plot_glass_brain(
                        img,
                        colorbar=True,
                        display_mode="ortho",
                        plot_abs=False,
                        figure=fig,
                        axes=ax,
                    )
                    fig.suptitle(
                        "Higher level nonparametric results (Displaying FWER-corrected Negative log10 p-values)"
                    )
                    # TODO: save the fig to proper location
                    # fig.savefig()
            else:
                if self.stat_map_masks is None:
                    masks = [None]
                    masks_names = [None]
                elif isinstance(self.stat_map_masks, list):
                    masks = self.stat_map_masks
                    masks_names = [f"mask{i+1}" for i in range(len(masks))]
                elif isinstance(self.stat_map_masks, dict):
                    masks = list(self.stat_map_masks.values())
                    masks_names = list(self.stat_map_masks.keys())
                else:
                    raise TypeError("stat")

                for mask, mask_name in zip(masks, masks_names):
                    if masks_names[0] is None:
                        plot_prefix = out_prefix + "_" + contrast_1level
                    else:
                        plot_prefix = (
                            out_prefix + "_" + contrast_1level + "_" + mask_name
                        )
                    mask = rbload_imgs(mask)[0]
                    self.plot_cluster_parametric(
                        returned_dict["z"],
                        mask_img=mask,
                        alphas=[0.05, 0.01, 0.001],
                        threshold=1.645,
                        cluster_threshold=0,
                        two_sided=True,
                        save=True,
                        prefix=plot_prefix,
                    )

    def loop_all_1level_contrasts(self):
        for contrast_1level, stat_maps in tqdm(self.stat_maps.items()):
            tqdm.write(f"Processing contrast {contrast_1level}")
            self.plot_all_stat(stat_maps, contrast_1level, save=True)
            out_prefix = contrast_1level
            stat_maps = rbload_imgs(stat_maps)
            group_results = self.process_single_1level_contrast(stat_maps, out_prefix)
            for contrast, imgs in group_results.items():
                self.higher_results[contrast].append(imgs)

        return self.higher_results

    def loop_all_2level_contrasts_parametric(self, fmri_glm, output_prefix):
        """
        THIS FUNCTION IS AN ALIAS OF loop_through_contrasts FOR CLARITY
        """
        higher_level_result = self.loop_all_contrasts_parametric(
            fmri_glm, output_prefix
        )

        return higher_level_result

    def loop_all_2level_contrasts_nonparametric(self, stat_maps, output_prefix):
        """
        References: https://nilearn.github.io/dev/modules/generated/nilearn.glm.second_level.non_parametric_inference.html#nilearn.glm.second_level.non_parametric_inference
        """
        return_dict = dict()
        if isinstance(self.contrasts, dict):
            conditions = self.contrasts.keys()
            con_matrix_all = self.contrasts.values()
        elif isinstance(self.contrasts, list):
            conditions = self.contrasts
            con_matrix_all = [None for i in range(len(conditions))]
        else:
            raise TypeError(
                "Contrasts must be a list of conditions or a dictionary of conditions and contrast matrices"
            )
        for condition, con_matrix in zip(conditions, con_matrix_all):
            if con_matrix is None:
                contrast = condition
            else:
                contrast = con_matrix
            if self.model_kwargs["tfce"] or self.model_kwargs["threshold"]:
                out_dic = non_parametric_inference(
                    stat_maps,
                    design_matrix=self.design_matrix,
                    second_level_contrast=contrast,
                    **self.model_kwargs,
                )
                return_dict[condition] = out_dic
                # self.plot_cluster_nonparametric(out_dic, output_prefix + "_" + condition)
                for key, value in out_dic.items():
                    value.to_filename(f"{output_prefix}_{condition}_{key}.nii.gz")
            else:
                neg_log10_vfwe_pvals_img = non_parametric_inference(
                    stat_maps,
                    design_matrix=self.design_matrix,
                    second_level_contrast=contrast,
                    **self.model_kwargs,
                )
                return_dict[condition] = neg_log10_vfwe_pvals_img
                # save img
                neg_log10_vfwe_pvals_img.to_filename(
                    f"{output_prefix}_{condition}_FWER-corrected_p.nii.gz"
                )

        return return_dict

    def plot_cluster_parametric(
        self,
        stat_img,
        mask_img=None,
        alphas: float or List[float] = None,
        threshold: float = 1.96,
        cluster_threshold: float = 10,
        two_sided: bool = True,
        save: bool = True,
        prefix: str = None,
    ):
        """
        Technically this function is exactly same as nilearn.glm.threshold_stats_img.
        This function loops all provided method (None|’fpr’|’fdr’|’bonferroni’) and different alpha level (0.05,0.01,0.001), then save a figure for convenience.

        Parameters
        ----------
        stat_img: Niimg-like object or None, optional
        Statistical image (presumably in z scale) whenever height_control is ‘fpr’ or None, stat_img=None is acceptable. If it is ‘fdr’ or ‘bonferroni’, an error is raised if stat_img is None.

        mask_img: Niimg-like object, optional,
        Mask image

        alphas : float or list, optional
        Number controlling the thresholding (either a p-value or q-value). Its actual meaning depends on the height_control parameter. This function translates alpha to a z-scale threshold. Default=0.001.

        threshold: float, optional
        Desired threshold in z-scale. This is used only if height_control is None. Default=3.0.

        cluster_threshold: float, optional
        cluster size threshold. In the returned thresholded map, sets of connected voxels (clusters) with size smaller than this number will be removed. Default=0.

        two_sided: Bool, optional
        Whether the thresholding should yield both positive and negative part of the maps. In that case, alpha is corrected by a factor of 2. Default=True.

        save:bool, optional
        If True, save all thresholded images to nii objects.

        prefix: str, optional
        Prefix for the outputfile. If save is True, prefix must be specified.

        Returns
        threshed_imgs_dict
        -------

        """
        if save:
            if not prefix:
                raise ValueError("If save is True, prefix must be specified.")
        threshed_imgs_dict = {}
        if isinstance(alphas, float):
            alphas = [alphas]
        elif alphas is None:
            alphas = [0.05, 0.01, 0.001]
        try:
            alphas = list(alphas)
        except:
            raise TypeError("alpha must be iterable if you didn't pass a single value")
        for alpha in alphas:
            threshed_imgs_dict[f"{alpha}"] = {}
            to_plot_tmp = {}
            for method in [None, "fpr", "fdr", "bonferroni"]:
                thresholded_map, threshold = threshold_stats_img(
                    stat_img=stat_img,
                    mask_img=mask_img,
                    alpha=alpha,
                    threshold=threshold,
                    height_control=method,
                    cluster_threshold=cluster_threshold,
                    two_sided=two_sided,
                )
                img_name = "None" if method is None else method
                threshed_imgs_dict[f"{alpha}"][img_name] = thresholded_map
                if save:
                    thresholded_map.to_filename(
                        self.out_dir
                        + f"{prefix}_{img_name}_threshed_zstat_alpha{alpha}.nii.gz"
                    )
                if img_name == "None":
                    to_plot_tmp[f"Z value threshold {threshold} "] = thresholded_map
                elif img_name == "fpr":
                    to_plot_tmp["False Positive Rate (FPR) Corrected"] = thresholded_map
                elif img_name == "fdr":
                    to_plot_tmp[
                        "False Discovery Rate (FDR) Corrected"
                    ] = thresholded_map
                elif img_name == "bonferroni":
                    to_plot_tmp["Bonferroni Corrected"] = thresholded_map

            fig, axes = plt.subplots(figsize=(8, 8), nrows=2, ncols=2)
            axes = axes.ravel()
            for img_counter, (title, img) in enumerate(to_plot_tmp.items()):
                if img_counter >= len(to_plot_tmp.keys()):
                    break
                ax = axes[img_counter]
                plot_glass_brain(
                    img,
                    colorbar=True,
                    display_mode="z",
                    plot_abs=False,
                    threshold=0,
                    figure=fig,
                    axes=ax,
                )
                ax.set_title(title)
            fig.suptitle(
                "Higher level parametric results (Displaying Thresholded Z-scaled statistics )"
            )
            fig.savefig(
                self.out_dir
                + f"{prefix}_higher_level_parametric_results_alpha{alpha}.png"
            )
            plt.close()

        return threshed_imgs_dict

    def plot_cluster_nonparametric(self, results: dict, output_prefix):
        threshold = -np.log(0.05)
        vmax = -np.log(1 / self.model_kwargs["n_perm"])

        images = []
        titles = []
        for img_name, img in results.items():
            img.to_filename(self.out_dir + f"{output_prefix}_{img_name}.nii.gz")
            if img_name in [
                "logp_max_t",
                "logp_max_size",
                "logp_max_mass",
                "logp_max_tfce",
            ]:
                images.append(results[img_name])
            if img_name == "logp_max_t":
                titles.append("Permutation Test\n(Voxel-Level Error Control)")
            elif img_name == "logp_max_size":
                titles.append("Permutation Test\n(Cluster-Size Error Control)")
            elif img_name == "logp_max_mass":
                titles.append("Permutation Test\n(Cluster-Mass Error Control)")
            elif img_name == "logp_max_tfce":
                titles.append("Permutation Test\n(TFCE Error Control)")

        fig, axes = plt.subplots(figsize=(8, 8), nrows=2, ncols=2)
        axes = axes.ravel()
        for img_counter, ax in enumerate(axes):
            if img_counter >= len(images):
                break
            plot_glass_brain(
                images[img_counter],
                colorbar=True,
                vmax=vmax,
                display_mode="z",
                plot_abs=False,
                threshold=threshold,
                figure=fig,
                axes=ax,
            )
            ax.set_title(titles[img_counter])
        fig.suptitle(
            "Higher level nonparametric results (Displaying Negative log10 p-values)"
        )
        fig.savefig(
            self.out_dir + f"{output_prefix}_higher_level_nonparametric_results.png"
        )
        plt.close()

    def plot_all_stat(
        self,
        list_stat_maps: list,
        out_name_prefix: str,
        plot_kwargs: dict = None,
        save: bool = True,
    ):
        fig, axes, cidx = None, None, None
        for idx, stat_map in enumerate(list_stat_maps):
            if idx % 25 == 0:
                if idx != 0 and save:
                    fig.savefig(
                        os.path.join(
                            self.out_dir, f"{out_name_prefix}_zmaps_{idx // 25}.png"
                        )
                    )
                    plt.close(fig)
                fig, axes = plt.subplots(nrows=5, ncols=5, figsize=(25, 25))
                axes = axes.ravel()
                cidx = 0

            plot_params = {
                "title": os.path.basename(stat_map),
                "axes": axes[cidx],
                "threshold": 1.96,
                "plot_abs": False,
                "display_mode": "z",
            }
            if plot_kwargs is not None:
                plot_params.update(plot_kwargs)
            self.plot_single_stat(
                stat_map, plot="glass", plot_kwargs=plot_params, save=False
            )
            cidx += 1

        if save and cidx > 0:  # Save the last figure if there are unsaved subplots
            fig.savefig(
                os.path.join(
                    self.out_dir, f"{out_name_prefix}_zmaps_{int((idx // 25) + 1)}.png"
                )
            )
            plt.close(fig)

    def fit(self, X, y=None):
        # X is expected to be a dictionary containing {contrast: [stat_maps]}
        self.stat_maps = X
        self.loop_all_1level_contrasts()

        return self

    def transform(self, X, y=None):
        # Return the `higher_results` attribute here,
        # as it contains the results of the second-level analysis
        return self.higher_results
